/*
 * Last Modified: 2007-09-05
 * Last Modified By: Jeff Huber
 * Original Author: Jamie Rasmussen
 *
 * Implements three Tcl commands:
 *  1) am_unquotehtml. Does the reverse of ns_quotehtml and am_quotehtml.
 *
 *  2) am_quotehtml. Analogous to ns_quotehtml except that it quotes double-quotes as &quot;
 *     and single quotes as &#39;
 *    (ns_quotehtml 2.x does not quote single or double-quotes
 *     at all and 3.x+ uses &#34; for double-quotes).
 *     Optional arguments that contain the strings "single" and/or "double" can be passed
 *      - If the 2nd or 3rd argument contains the string "double", double-quotes will NOT be converted
 *      - If the 2nd or 3rd argument contains the string "single", single-quotes will NOT be converted
 *  
 *  3) am_attribute2set which takes a string like
 *     {a=b c='d' e="f" selected} and turns it into an ns_set.  This is the
 *     logical inverse of our tcl am_set2attributes proc.  By default, we
 *     handle boolean attributes (see 
 *     http://www.w3.org/TR/REC-html40/intro/sgmltut.html#h-3.3.4.2) the
 *     same way AOLserver 4 does when creating an ns_set of attributes to 
 *     pass to the proc registered for a fancy adp tag; that is, we set the
 *     key to the value.  If a switch that starts with "-minimize" is passed,
 *     we instead use a zero-length key. This is to allow round-tripping such an
 *     ns_set back to attributes using am_set2attributes with bypass in onEmptyKey.
 *     If a switch that starts with "-unquote" is passed we html-unquote the values.
 *
 *     The ParseAtts function is a slightly modified version of the one in
 *     AOLserver 4's adpparse.c.
 */

/*
 * The contents of this file are subject to the AOLserver Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://aolserver.com/.
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is AOLserver Code and related documentation
 * distributed by AOL.
 * 
 * The Initial Developer of the Original Code is America Online,
 * Inc. Portions created by AOL are Copyright (C) 1999 America Online,
 * Inc. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public License (the "GPL"), in which case the
 * provisions of GPL are applicable instead of those above.  If you wish
 * to allow use of your version of this file only under the terms of the
 * GPL and not to allow others to use your version of this file under the
 * License, indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by the GPL.
 * If you do not delete the provisions above, a recipient may use your
 * version of this file under either the License or the GPL.
 */


//static const char *RCSID = "@(#) $Header: /cvsroot/aolserver/amattributes2set/amattributes2set.c,v 1.1 2000/08/20 19:12:26 kriston Exp $, compiled: " __DATE__ " " __TIME__;

#include "ns.h"
//#include "amattributes2set.h"

int Ns_ModuleVersion = 1;

int Ns_ModuleInit(char *hServer, char *hModule);
static int ParseInterpInit(Tcl_Interp *interp, void *context);

// Adds am_attribute2set
static int NsTclAttributesToSetObjCmd(ClientData clientdata, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);
static void ParseAtts(char *s, char *e, Ns_Set* set, int minimize,  short int unquote);

// Adds am_unquotehtml
static void Am_UnQuoteHtml(Ns_DString *pds, CONST84 char *string);
static Tcl_CmdProc AmUnQuoteHtmlCmd;

// Adds am_quotehtml
static void Am_QuoteHtml(Ns_DString *pds, char *string, short doNotQuoteDoubleQuotes, short doNotQuoteSingleQuotes);
static Tcl_CmdProc AmQuoteHtmlCmd;


/*
 *----------------------------------------------------------------------
 *
 * Ns_ModuleInit --
 *
 *      This is the amattributes2set module's entry point.  AOLserver runs
 *      this function right after the module is loaded.  It is used to
 *      read configuration data, initialize data structures, kick off
 *      the Tcl initialization function (if any), and do other things
 *      at startup.
 *
 * Results:
 *      NS_OK or NS_ERROR
 *
 * Side effects:
 *      Module loads and initializes itself.
 *
 *----------------------------------------------------------------------
 */

int
Ns_ModuleInit(char *hServer, char *hModule)
{
    return (Ns_TclInitInterps(hServer, ParseInterpInit, NULL));
}


/*
 *----------------------------------------------------------------------
 *
 * ParseInterpInit --
 *
 *      Register new commands with the Tcl interpreter.
 *
 * Results:
 *      NS_OK or NS_ERROR
 *
 * Side effects:
 *      A C function is registered with the Tcl interpreter.
 *
 *----------------------------------------------------------------------
 */

static int
ParseInterpInit(Tcl_Interp *interp, void *context)
{
    Tcl_CreateObjCommand(interp, "am_attributes2set", NsTclAttributesToSetObjCmd, NULL, NULL);

	Tcl_CreateCommand(interp, "am_unquotehtml", AmUnQuoteHtmlCmd,
		(ClientData) NULL, NULL);

	Tcl_CreateCommand(interp, "am_quotehtml", AmQuoteHtmlCmd,
		(ClientData) NULL, NULL);

    return NS_OK;
}


static int
NsTclAttributesToSetObjCmd( ClientData clientdata, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] ) {
    Ns_Set *set;
    char *input;
    int size;
    int minimize = 0, unquoteHtml = 0;

    if (objc == 2) {
        input = Tcl_GetStringFromObj(objv[1], &size);
    } else if (objc == 3 || objc == 4) {
        int s, i;
        char * opt;

        // Iterate thru the number of parameters possible (in this case two)
        for (i = 1; i <= objc - 2; i++) {
            opt = Tcl_GetStringFromObj(objv[i], &s);
            if (s <= 1) {
                Tcl_AppendStringsToObj(Tcl_GetObjResult(interp), "bad option \"",
                    opt, "\": must be -minimize and/or -unquote", (char *) NULL);
                return TCL_ERROR;
            }

            if (strstr(opt, "-minimize") != '\0')
                minimize = 1;
            else if (strstr(opt, "-unquote") != '\0')
                unquoteHtml = 1;
            else {
                Tcl_AppendStringsToObj(Tcl_GetObjResult(interp), "bad option \"",
                    opt, "\": must be -minimize or -unquote", (char *) NULL);
                return TCL_ERROR;            
            }

        }
 
        input = Tcl_GetStringFromObj(objv[objc-1], &size);
    } else {
        Tcl_WrongNumArgs(interp, 1, objv, "?-minimize? ?-unquote?");
        return TCL_ERROR;
    }

    set = Ns_SetCreate(NULL);
    ParseAtts(input, input + size, set, minimize, unquoteHtml);
    Ns_TclEnterSet(interp, set, NS_TCL_SET_DYNAMIC);
    return NS_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * ParseAtts --
 *
 *	Parse tag attributes, copying cleaned up pairs to given ns_set.
 *
 *----------------------------------------------------------------------
 */

static void
ParseAtts(char *s, char *e, Ns_Set* set, int minimize, short int unquote)
{
    char *vs = NULL, *ve = NULL, *as = NULL, *ae = NULL;
    char end = 0, vsave = 0, asave = 0, foundBoolAttr = 0;
    
    // If unquote parameter is true, this variable is used
    // to temporarily store the value html-unquoted. This function
    // could be written to avoid these temporary variables, but
    // it'd be time consuming to do and the loss of efficiency by not
    // doing so isn't that critical (for now).
    Ns_DString unquotedValue;
    char *valueToPut;

    if (unquote == 1) 
        Ns_DStringInit(&unquotedValue);

    while (s < e) {
        /*
        * Trim attribute name.
        */

        while (s < e && isspace(UCHAR(*s))) {
            ++s;
        }
        if (s == e) {
            break;
        }
        as = s;

        if (*s != '\'' && *s != '"') {
            while (s < e && !isspace(UCHAR(*s)) && *s != '=') {
                ++s;
            }
        } else {
            ++s;
            while (s < e && *s != *as) {
                ++s;
            }
            ++s;
        }

        ae = s;
        while (s < e && isspace(UCHAR(*s))) {
            ++s;
        }
        if (*s != '=') {
            /*
            * Use attribute string as both name and value
            */

            vs = as;
            foundBoolAttr = 1;
        } else {
            /*
            * Trim spaces and/or quotes from value.
            */

            do {
                ++s;
            } while (s < e && isspace(UCHAR(*s)));
            vs = s;

            if (*s != '"' && *s != '\'') {
                while (s < e && !isspace(UCHAR(*s))) {
                    ++s;
                }
            } else {
                ++s;
                while (s < e && *s != *vs) {
                    ++s;
                }
                ++s;
            }

            ve = s;
            end = *vs;
            if (end != '=' && end != '"' && end != '\'') {
                end = 0;
            }
            if (end && ve > vs && ve[-1] == end) {
                ++vs;
                --ve;
            }
            vsave = *ve;
            *ve = '\0';
        }
        asave = *ae;
        *ae = '\0';

        if (unquote == 1) {
            // If unquote is specified, unquote the string
            // and specify the unquoted string to be 
            // put in the set
            Am_UnQuoteHtml(&unquotedValue, vs);
            valueToPut = Ns_DStringValue(&unquotedValue);
        } else {
            // Unquote isn't specified. Put the raw value
            valueToPut = vs;
        }

        if (minimize && foundBoolAttr) {
            Ns_SetPut(set, "", valueToPut);
        } else {
            Ns_SetPut(set, as, valueToPut);
        }

        /* 
        * Restore strings.
        */

        if (unquote == 1) {
            Ns_DStringTrunc(&unquotedValue, 0);
        }

        *ae = asave;
        if (vs != as) {
            *ve = vsave;
        }
        foundBoolAttr = 0;
    }

    if (unquote == 1) {
        Ns_DStringFree(&unquotedValue);
    }
}

/*
*----------------------------------------------------------------------
*
* AmUnQuoteHtmlCmd --
*
*	Implements am_unquotehtml. Does the reverse of ns_quotehtml. 
*
* Results:
*	Tcl result. 
*
* Side effects:
*	See docs. 
*
*----------------------------------------------------------------------
*/
static int AmUnQuoteHtmlCmd(ClientData string, Tcl_Interp *interp, int argc, CONST84 char *argv[]) {

    Ns_DString ds;

    if (argc != 2) {
        Tcl_AppendResult(interp, "wrong # args:  should be \"",
                         argv[0], " string\"", NULL);
        return TCL_ERROR;
    }
    
    Ns_DStringInit(&ds);

    Am_UnQuoteHtml(&ds, argv[1]);
    
    Tcl_SetResult(interp, Ns_DStringExport(&ds), (Tcl_FreeProc *) ns_free);
        
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Am_UnQuoteHtml --
 *
 *	UnQuote an HTML string.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Copies unquoted HTML to given dstring. Based on Ns_QuoteHtml.
 *
 *----------------------------------------------------------------------
 */

static void
Am_UnQuoteHtml(Ns_DString *pds, CONST84 char *string)
{
    while (*string != '\0') {

         if (strncmp(string, "&amp;", 5) == 0) {
            Ns_DStringAppend(pds, "&");            
            string += 5;

         } else if (strncmp(string, "&#39;", 5) == 0) {
            Ns_DStringAppend(pds, "'");    
            string += 5;

         } else if (strncmp(string, "&quot;", 6) == 0) {
            Ns_DStringAppend(pds, "\"");  
            string += 6;

         } else if (strncmp(string, "&#34;", 5) == 0) {
            Ns_DStringAppend(pds, "\"");  
            string += 5;

         } else if (strncmp(string, "&lt;", 4) == 0) {
            Ns_DStringAppend(pds, "<");  
            string += 4;

         } else if (strncmp(string, "&gt;", 4) == 0) {
            Ns_DStringAppend(pds, ">");  
            string += 4;

         } else {
            Ns_DStringNAppend(pds, string, 1); 
            ++string;
         }
    }

}


/*
 *----------------------------------------------------------------------
 *
 * AmQuoteHtmlCmd --
 *
 *	Implements am_quotehtml. 
 *
 * Results:
 *	Tcl result. 
 *
 * Side effects:
 *	See docs. 
 *
 *----------------------------------------------------------------------
 */

static int
AmQuoteHtmlCmd(ClientData dummy, Tcl_Interp *interp, int argc, char **argv)
{
    Ns_DString ds;
    short doNotQuoteDoubleQuotes = 0, doNotQuoteSingleQuotes = 0;
    int i;

    if (argc < 2 || argc > 4) {
        Tcl_AppendResult(interp, "wrong # args:  should be \"",
            argv[0], " html ?single? ?double?\"", NULL);
        return TCL_ERROR;
    }

    if (argc == 3 || argc == 4) {
        for (i = 2; i < argc; i++) {
            if (strstr(argv[i], "single") != '\0') {
                doNotQuoteSingleQuotes = 1;
            }

            if (strstr(argv[i], "double") != '\0') {
                doNotQuoteDoubleQuotes = 1;
            }
        }
    }

    Ns_DStringInit(&ds);
    
    Am_QuoteHtml(&ds, argv[1], doNotQuoteDoubleQuotes, doNotQuoteSingleQuotes);
    
    Tcl_SetResult(interp, Ns_DStringExport(&ds), (Tcl_FreeProc *) ns_free);
    
    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * Am_QuoteHtml --
 *
 *	Quote an HTML string. Based on Ns_QuoteHtml, but changes "s to &quot; instead of
 *  &#34;. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Copies quoted HTML to given dstring.
 *
 *----------------------------------------------------------------------
 */

static void
Am_QuoteHtml(Ns_DString *pds, char *string, short doNotQuoteDoubleQuotes, short doNotQuoteSingleQuotes)
{
    while (*string != '\0') {
        switch (*string) {
        case '<':
            Ns_DStringAppend(pds, "&lt;");
            break;

        case '>':
            Ns_DStringAppend(pds, "&gt;");
            break;

        case '&':
            Ns_DStringAppend(pds, "&amp;");
            break;

        case '\'':
            if (doNotQuoteSingleQuotes != 1)
                Ns_DStringAppend(pds, "&#39;");
            else
                Ns_DStringNAppend(pds, string, 1);

            break;

        case '"':
            if (doNotQuoteDoubleQuotes != 1) 
                Ns_DStringAppend(pds, "&quot;");
            else
                Ns_DStringNAppend(pds, string, 1);

            break;

        default:
            Ns_DStringNAppend(pds, string, 1);
            break;
        }
        ++string;
    }
}
